
function log_likelihood = noisy_or_tree_likelihood( parent_ix, local_evidence, q, truth )
% Efficiently computes the likelihood of the data given local evidence.
% 
% David Duvenaud
% March 2011
%
% Inputs:
%     parent_ix:  the tree structure, given as a list of parent pointers.
%                  e.g. parent_ix = [ 0 1 1 2 2 3 3 4 4 5 5 6 6 7 7];
%     local_evidence:  the prior probability of each node being on [0 to 1]
%     q:  the noisy-or failure rate.
%     truth:  The setting of the variables whose likelihood will be
%             evaluated.

    % Precache the children of each node as a cell array.
    all_children = cell( length(parent_ix), 1 );
    num_factors = 0;
    for i = 1:length( parent_ix )
        all_children{i} = find( parent_ix == i );        
        if ~isempty( all_children{i} )
            num_factors = num_factors + 1;
        end
    end

    root = find( parent_ix == 0 );  
    
    % get conditional and unconditional probability of the root being on,
    % plus the normalization constant
    [ p_cond_off, p_uncond_off, log_p_child_truth ] = recursive_likelihood( root, parent_ix, local_evidence, truth, q, all_children );
        
    % combine with local evidence and truth
    log_cond_sum = log(( local_evidence( root ) * truth(root) * ( 1 - p_cond_off ) + ( 1 - local_evidence( root )) * ( 1 - truth(root)) * p_cond_off ));    
    log_uncond_sum = log(local_evidence( root ) * ( 1 - p_uncond_off ) + ( 1 - local_evidence( root )) * p_uncond_off);    
    log_likelihood = log_cond_sum - log_uncond_sum + log_p_child_truth;    
end


function [ p_cond_off, p_uncond_off, log_p_child_truth ] = recursive_likelihood( cur_node, parent_ix, local_evidence, truth, q, all_children )
    
cur_children = all_children{cur_node};

if isempty( cur_children )
    % if leaf
    p_cond_off = 0.5;
    p_uncond_off = 0.5;
    log_p_child_truth = 0;
else
    % get marginals over all children
    num_children = length( cur_children );
    child_p_cond_on = NaN( num_children, 1 );  
    child_p_cond_off = NaN( num_children, 1 );
    child_p_uncond_on = NaN( num_children, 1 );  
    child_p_uncond_off = NaN( num_children, 1 );
    
    log_p_child_truth = 0;
    for c = 1:num_children
        
        % recusively get the children's likelihoods and logZ
        [ child_p_cond_off(c), child_p_uncond_off(c), cur_log_child_p_child_truth ] = recursive_likelihood( cur_children(c), parent_ix, local_evidence, truth, q, all_children );           
        child_p_cond_on(c) = 1 - child_p_cond_off(c);
        child_p_uncond_on(c) = 1 - child_p_uncond_off(c);
        log_p_child_truth = log_p_child_truth + cur_log_child_p_child_truth;
    end
    
    % compute several sums of products
    % ===========================================    
    %p_cond_off = 0;   % P( y_p = 0, y1 ... yc in truth | e )
    %p_uncond_off = 0; % P( y_p = 0 | e )
    %p_truth = 0;      % P( y1 ... yc in truth )
    %p_ev = 0;         % P( y1 ... yc )
    
    z = 1:num_children;
    i = cur_children(z);
    log_p_cond_off = sum( log(  truth(i) .* local_evidence(i) .* child_p_cond_on(z)   .* q + ( 1 - truth(i)) .* ( 1 - local_evidence(i)) .* child_p_cond_off(z) ));
    log_p_uncond_off = sum(log(             local_evidence(i) .* child_p_uncond_on(z) .* q +                    ( 1 - local_evidence(i)) .* child_p_uncond_off(z) ));        
    log_p_truth = sum( log(     truth(i) .* local_evidence(i) .* child_p_cond_on(z)        + ( 1 - truth(i)) .* ( 1 - local_evidence(i)) .* child_p_cond_off(z) ));        
    log_p_ev = sum( log(                    local_evidence(i) .* child_p_uncond_on(z)      +                    ( 1 - local_evidence(i)) .* child_p_uncond_off(z) ));   
    
    % normalize
    p_uncond_off = exp(log_p_uncond_off - log_p_ev);
    
    % normalize but avoid 0 / 0
    if isinf( log_p_truth )
        p_cond_off = 0;
        log_p_child_truth = -Inf;
    else
        p_cond_off = exp(log_p_cond_off - log_p_truth);
        log_p_child_truth = log_p_child_truth + log_p_truth - log_p_ev;
    end
end
end

