
function marginals = noisy_or_tree_marginals( parent_ix, local_evidence, q )
% Uses BP to compute the posterior marginals on a set of tree with 
% noisy-or factors.  Optionally computes the likelihood of the data.

% Inputs:
%     parent_ix:  the tree structure, given as a list of parent pointers.
%                  e.g. parent_ix = [ 0 1 1 2 2 3 3 4 4 5 5 6 6 7 7];
%     local_evidence:  the prior probability of each node being on [0 to 1]
%     q:  the noisy-or failure rate.

    if nargin < 3
        q = 0.1;
    end

    %precache children
    all_children = cell( length(parent_ix), 1 );
    num_factors = 0;
    for i = 1:length( parent_ix )
        all_children{i} = find( parent_ix == i );        
        if ~isempty( all_children{i} )
            num_factors = num_factors + 1;
        end
    end

    % upward pass
    messages_sideways = ones( size(parent_ix ) ) .* 0.5;
    messages_up = non_recursive_pass_up( parent_ix, all_children, local_evidence, messages_sideways, q );           

    % downward pass
    marginals = nonrecursive_pass_down( all_children, parent_ix, local_evidence, messages_up, q );
end



function messages_up = non_recursive_pass_up( parent_ix, all_children, local_evidence, messages_sideways, q )
    
    messages_up = NaN( size(parent_ix ));
    
    % find all leaves
    isleaf = zeros( length(all_children), 1 );
    for i = 1:length(all_children)
        isleaf(i) = isempty(all_children{i});
    end
    leaves = find(isleaf);
    
    % leaf nodes   
    % =============
    parents = NaN(length(leaves), 1);
    for i = 1:length(leaves)
        cur_node = leaves(i);
        
        % just return the local evidence of being on times sideways messages     
        prob_off = ( 1 - local_evidence( cur_node )) * ( 1 - messages_sideways( cur_node ));
        prob_on = ( local_evidence( cur_node )) * messages_sideways( cur_node );        
        prob = prob_on / ( prob_off + prob_on );   % normalize 
        
        if isnan(prob)
            fprintf('~');
            prob = 0.5;
        end
        
        messages_up( cur_node ) = prob;
        parents(i) = parent_ix( cur_node );
    end
    parents = unique(parents );  % slow!
        
    % parent nodes
    % ==================
    while parents ~= 0 
        
        newparents = NaN(length(parents), 1);
        
        for i = 1:length(parents)

            cur_node = parents(i);
            children = all_children{cur_node};
            
            % check if we've computed all our children's messages
            if any( isnan( messages_up( children )))
                % compute this one next time
                newparents(i) = cur_node;
            else

                % get our children's messages
                child_probs = messages_up( children );

                % combine them all        
                OR_product = prod( 1 - (1 - q) .* child_probs );

                % add local evidence
                lambda = local_evidence( cur_node );  % if we want to enable sideways messages at higher levels, add it here
                prob_off = ( 1 - lambda ) * OR_product;      % noisy-or CPD - equations 10 and 11
                prob_on = lambda * ( 1 - OR_product );
                prob = prob_on / ( prob_on + prob_off );

                % numerical stability hack
                if isnan( prob ) || isinf( prob )
                    prob = 1;
                    fprintf('!');
                end

                messages_up( cur_node ) = prob;

                newparents(i) = parent_ix( cur_node );
            end
        end
        parents = unique(newparents);
    end 
end


function [ marginals, messages_down ] = nonrecursive_pass_down( all_children, parent_ix, local_evidence, messages_up, q )  
    
    roots = find( parent_ix == 0 );    
    cur_set = roots;
    
    marginals = NaN( size( parent_ix));
    messages_down = NaN( size( parent_ix));
    messages_down(roots) = 0.5;
    
    while ~isempty(cur_set)

        % compute marginals
        for i = 1:length(cur_set)
            cur_node = cur_set(i);
            
            % find our marginal by combining messages down and messages up.
            % Local evidence is already in the message up.
            prob_off = ( 1 - messages_down( cur_node )) * ( 1 - messages_up( cur_node ));
            prob_on = ( messages_down( cur_node )) * messages_up( cur_node );        
            prob = prob_on / ( prob_off + prob_on );   % normalize
            
            if isnan( prob  )
                prob = 0.5;
                fprintf('z');
            end
            
            marginals( cur_node ) = prob;
        end
        
        % compute messages down
        next_set2 = cell( length(cur_set), 1 );
        
        for i = 1:length(cur_set)
            cur_node = cur_set(i);
            children = all_children{cur_node};
            
            % change this if we want to use non-regular tree topologies
            if 0%isempty(children)
                
                % we're at the bottom, so we don't have to compute any more
                % messages down!
                return;                
            else
                % note who our children are for the next iteration
                next_set2{i} = children;
                
                % todo: vectorize this?
                for c = 1:length(children)

                    % compute pi_prime
                    cur_ix = children;
                    cur_ix( c ) = []; % remove yourself from the list            
                    pi_c = messages_up( cur_ix );    % note: this is slower than it needs to be                   
                    pi_prime = prod( 1 - (1 - q) * pi_c );  % product of everyone but yourself                
                    % todo: rearrange the above to be more stable and faster.  maybe do
                    % a sigmoid transform?                

                    % compute lambda
                    lambda_on = local_evidence( cur_node ) * messages_down( cur_node );
                    lambda_off = ( 1 - local_evidence( cur_node )) * ( 1 - messages_down( cur_node ));
                    lambda = lambda_on / ( lambda_on + lambda_off );

                    % compute downwards message
                    down_prob_on = ( 1 - lambda ) * q * pi_prime + lambda * ( 1 - q * pi_prime );
                    down_prob_off = ( 1 - lambda ) * pi_prime + lambda * ( 1 - pi_prime );                    
                    down_prob = down_prob_on / ( down_prob_off + down_prob_on );   % normalize
                    messages_down(children(c)) = down_prob;
                    assert( ~isnan(down_prob));
                end      
            end                
        end
        
        % flatten next set
        cur_set = [];
        for i = 1:length( next_set2 )
            cur_set = [ cur_set next_set2{i} ];
        end
    end
end

